-----------------------------------
-----------------------------------
-- SpamSentry by Anea
-----------------------------------
-- report.lua
-- Spam management and report functions
-----------------------------------

SS_Report = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0")
local L = AceLibrary("AceLocale-2.2"):new("SS_Report")

-----------------------------------
-----------------------------------
-- Initialisation

setglobal("SSGUIHELP", L["SSGUIHELP"])
setglobal("SSGUISEND", L["Send"])
setglobal("SSGUIRESET", L["Reset"])
setglobal("SSGUICLOSE", L["Close"])

function SS_Report:OnEnable()
  -- Setup the default ticket texts
  local s = SS.dbr.realm.spamTicketText
  local b = SS.dbr.realm.botTicketText
  local r = SS.dbr.realm.rpTicketText
  if not s or s =="" then
    SS.dbr.realm.spamTicketText = getglobal("SPAMSENTRY_SPAMTEXT_"..SS.dbr.realm.reportLanguage)
  end
  if not b or b =="" then
    SS.dbr.realm.botTicketText = getglobal("SPAMSENTRY_BOTTEXT_"..SS.dbr.realm.reportLanguage)
  end
  if not r or r =="" then
    SS.dbr.realm.rpTicketText = getglobal("SPAMSENTRY_RPTEXT_"..SS.dbr.realm.reportLanguage)
  end
end

-----------------------------------
-----------------------------------
-- Spam list and edit functions

function SS_Report:ListSpam()
  local l = SS.dbr.realm.spamReportList
  local num = getn(l)
  if num>0 then
    SS:Msg(0, L["Blocked messages:"])
    for i=1, num, 1 do
      character = l[i].player
      message = gsub(l[i].message, "\n", " ")
      typ = l[i].type;
      if strlen(message)>50 then
        message = strsub(message, 1, 50)
      end
      SS:Msg(0, format("%s. %s(%s): |cffffffff%s ...|r" , tostring(i), SS:PlayerLink(character), typ, message))
    end
  else
    SS:Msg(0, L["No blocked messages"])
  end
end

function SS_Report:RemoveSpam(name, manual)
  local l = SS.dbr.realm.spamReportList
  local index = SS:InList(l, name)
  if not index then return end
  if SS.chatHistory[name] then -- Reset the chatcache to prevent re-blocking
    SS.chatHistory[name].message = {}
    SS.chatHistory[name].time = {}
  end
  tremove(SS.dbr.realm.spamReportList, index)
  if manual then
    DelIgnore(name)
    tremove(SS.characterBlackList, index)
  end
  SS:TriggerEvent("SPAMSENTRY_REPORTLIST_UPDATED")
  SS:Msg(0, format(L["%s has been removed"], name))  
end

function SS_Report:ClearSpam(num, manual)
  if num==0 then
    num = getn(SS.dbr.realm.spamReportList)
  end
  for i=1, num, 1 do
    local name = SS.dbr.realm.spamReportList[1].player
    if SS.chatHistory[name] then -- Reset the chatcache to prevent re-blocking
      SS.chatHistory[name].message = {}
      SS.chatHistory[name].time = {}
    end
    tremove(SS.dbr.realm.spamReportList, 1)
    if manual then
      DelIgnore(name)
      tremove(SS.characterBlackList, 1)
    end
  end
  SS:TriggerEvent("SPAMSENTRY_REPORTLIST_UPDATED")
  SS:Msg(0, format(L["Reportlist cleared (%s items)"], num))
end

-----------------------------------
-----------------------------------
-- Ticket functions

-- Show the SpamSentry ticketframe GUI
function SS_Report:ShowGUI(t)
  SpamSentry_ButtonReset:Hide()
  SpamSentry_ButtonSend:Hide()
  SpamSentry_ButtonEdit:Hide()
  SpamSentry_ButtonResetBot:Hide()
  SpamSentry_ButtonSendBot:Hide()
  SpamSentry_ButtonEditBot:Hide()
  SpamSentry_ButtonResetRP:Hide()
  SpamSentry_ButtonSendRP:Hide()
  SpamSentry_ButtonEditRP:Hide()

  if t=="spam" then
    SpamSentry_HelpText:SetText(L["SSGUIHELP"])
    SpamSentry_ButtonReset:Show()
    SpamSentry_ButtonSend:Show()
    SpamSentry_Text:SetText(self:MakeSpamReport())
  elseif t=="bot" then
    SpamSentry_HelpText:SetText(L["SSGUIHELP"])
    SpamSentry_ButtonResetBot:Show()
    SpamSentry_ButtonSendBot:Show()
    SpamSentry_Text:SetText(self:MakeBotReport())
  elseif t=="rp" then
    SpamSentry_HelpText:SetText(L["SSGUIHELP"])
    SpamSentry_ButtonResetRP:Show()
    SpamSentry_ButtonSendRP:Show()
    SpamSentry_Text:SetText(self:MakeRPReport())
  elseif t=="editspam" then
    SpamSentry_HelpText:SetText(L["SSGUIHELP_EDITTICKETTEXT"])
    SpamSentry_ButtonEdit:Show()
    SpamSentry_Text:SetText(SS.dbr.realm.spamTicketText)
  elseif t=="editbot" then
    SpamSentry_HelpText:SetText(L["SSGUIHELP_EDITTICKETTEXT"])
    SpamSentry_ButtonEditBot:Show()
    SpamSentry_Text:SetText(SS.dbr.realm.botTicketText)
  elseif t=="editrp" then
    SpamSentry_HelpText:SetText(L["SSGUIHELP_EDITTICKETTEXT"])
    SpamSentry_ButtonEditRP:Show()
    SpamSentry_Text:SetText(SS.dbr.realm.rpTicketText)
  elseif t=="feedback" then
    SpamSentry_HelpText:SetText(L["SSGUIHELP_FEEDBACK"])
    SpamSentry_Text:SetText(self:FeedBackText())
  end
  SpamSentryGUI:Show()
end

function SS_Report:OpenTicket(reptype)
  if PETITION_QUEUE_ACTIVE == nil then
    SS:Msg(0, HELP_TICKET_QUEUE_DISABLED)
    return
  end
  local t = getglobal("SpamSentry_Text"):GetText()
  if not HelpFrameOpenTicket.hasTicket then
    if reptype == "spam" then
      NewGMTicket(2, t)
      self:ClearSpam(SS.db.profile.singlePlayerTicket and 1 or min(4, getn(SS.dbr.realm.spamReportList)))
    elseif reptype == "bot" then
      NewGMTicket(2, t)
      SS_Bot:Clear(SS.db.profile.singlePlayerTicket and 1 or min(4, getn(SS.dbr.realm.botReportList)))
    elseif reptype == "rp" then
      NewGMTicket(2, t)
      SS_RP:Clear(SS.db.profile.singlePlayerTicket and 1 or min(4, getn(SS.dbr.realm.rpReportList)))
    end
  else
    SS:Msg(0,L["You already have a ticket pending"])
  end
end

-- Create the spam report-text
function SS_Report:MakeSpamReport()
  local text = L["Nothing to report"]
  local report = "\n"
  local l = SS.dbr.realm.spamReportList
  local num = SS.db.profile.singlePlayerTicket and 1 or min(4, getn(l))   -- Limit reports to 5 due to 500-char ticket limit.

  if num >0 then
    -- Create extended report if there is only 1 character to report
    if num == 1 then
      report = report .. format("%s - %s (%s):\n%s\n\n", l[1].time, l[1].player, l[1].type, l[1].message)
    -- Create brief list if there are 2-5 characters to report
    else
      report = L["time / player / type / company|message"].."\n"
      for i=1, num, 1 do
        report = report .. format("%s - %s (%s): %s\n", l[i].time, l[i].player, l[i].type, self:FindCompany(l[i].message))
      end
    end
    text = format(SS.dbr.realm.spamTicketText, UnitName("player"))..report
  end
  return text,num
end

-- Create the bot report-text
function SS_Report:MakeBotReport()
  local text = L["Nothing to report"]
  local report = "\n"
  local l = SS.dbr.realm.botReportList
  local num = SS.db.profile.singlePlayerTicket and 1 or min(5, getn(l))   -- Limit reports to 5 due to the 500-char ticket limit.

  if num > 0 then
    for i=1, num, 1 do
      raceclass = ""
      if l[i].race and l[i].class then
        raceclass = "("..l[i].race.." "..l[i].class..")"
      end
      report = report .. format("%s %s- %s - %s. "..L["Farming:"].." %s\n", l[i].player, raceclass, l[i].date, l[i].location, l[i].mob or "")
    end
    text = format(SS.dbr.realm.botTicketText, UnitName("player"))..report
  end
  return text,count
end

-- Create the RP report-text
function SS_Report:MakeRPReport()
  local text = L["Nothing to report"]
  local report = "\n"
  local l = SS.dbr.realm.rpReportList
  local num = SS.db.profile.singlePlayerTicket and 1 or min(5, getn(l))   -- Limit reports to 5 due to the 500-char ticket limit.

  if num > 0 then
    for i=1, num, 1 do
      raceclass = ""
      if l[i].race and l[i].class then
        raceclass = l[i].race.." "..l[i].class
        if not raceclass=="" then raceclass = "("..raceclass..")" end
      end
      report = report .. format("%s %s- %s\n", l[i].player, raceclass, l[i].date)
    end
    text = format(SS.dbr.realm.rpTicketText, UnitName("player"))..report
  end
  return text,count
end

-- Returns the company-name or a summary of the message
function SS_Report:FindCompany(msg)
  local tmsg = string.lower(gsub(msg, SS.strip, ""))
  l = SS.websites
 
  for i,v in pairs(SS.websites) do
    if strfind(tmsg, i) then
      return v
    end
  end

  if strlen(msg)>30 then
    msg = strsub(msg, 1, 30)
  end
  msg = gsub(msg, "\n", " ")
  return msg
end

-----------------------------------
-----------------------------------
-- Ticket text editing

function SS_Report:ResetTicketText()
  SS.dbr.realm.spamTicketText = getglobal("SPAMSENTRY_SPAMTEXT_"..SS.dbr.realm.reportLanguage)
  SS.dbr.realm.botTicketText = getglobal("SPAMSENTRY_BOTTEXT_"..SS.dbr.realm.reportLanguage)
  SS.dbr.realm.rpTicketText = getglobal("SPAMSENTRY_RPTEXT_"..SS.dbr.realm.reportLanguage)
end

-----------------------------------
-----------------------------------
-- Feedback functions

function SS_Report:FeedBackText()
  local l = SS.dbr.realm.spamReportList
  local index = SS:InList(l, SS.spamIndex)
  local ret = ""
  if index then
    return self:StringToHex(l[index].message)
  end
  return ""
end

local str2hex = {"0","1","2","3","4","5","6","7","8","9","A","B","C","D","E","F"}
function SS_Report:StringToHex(s)
  local ret, b, n, m = "", 0, 0 ,0
  for i=1, strlen(s), 1 do
    b = strbyte(s, i)
    n = mod(b, 16)
    m = (b - n) / 16
    ret = ret..str2hex[m+1]..str2hex[n+1]
  end
  return ret
end
